"use strict";
cc._RF.push(module, '32079EnDKxLcb7nVjYmHoTS', 'protocol');
// Script/pomelo/protocol/protocol.ts

"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var buffer_1 = require("../buffer/buffer");
var PKG_HEAD_BYTES = 4;
var HASH_BYTES = 4;
var END_MSGID_BYTES = 4;
var MSG_FLAG_BYTES = 1;
var MSG_ROUTE_CODE_BYTES = 2;
var MSG_ID_MAX_BYTES = 5;
var MSG_ROUTE_LEN_BYTES = 1;
var MSG_ROUTE_CODE_MAX = 0xffff;
var MSG_COMPRESS_ROUTE_MASK = 0x1;
var MSG_COMPRESS_GZIP_MASK = 0x1;
var MSG_COMPRESS_GZIP_ENCODE_MASK = 1 << 4;
var MSG_TYPE_MASK = 0x7;
var hasSeqIdEncode = true; //是否使用msgid
var hasSeqIdDecode = false; //是否使用msgid
var SEED = 9527;
var Crc32;
(function (Crc32) {
    var T = signed_crc_table();
    function signed_crc_table() {
        var c = 0, table /*:Array<number>*/ = new Array(256);
        for (var n = 0; n != 256; ++n) {
            c = n;
            c = ((c & 1) ? (-306674912 ^ (c >>> 1)) : (c >>> 1));
            c = ((c & 1) ? (-306674912 ^ (c >>> 1)) : (c >>> 1));
            c = ((c & 1) ? (-306674912 ^ (c >>> 1)) : (c >>> 1));
            c = ((c & 1) ? (-306674912 ^ (c >>> 1)) : (c >>> 1));
            c = ((c & 1) ? (-306674912 ^ (c >>> 1)) : (c >>> 1));
            c = ((c & 1) ? (-306674912 ^ (c >>> 1)) : (c >>> 1));
            c = ((c & 1) ? (-306674912 ^ (c >>> 1)) : (c >>> 1));
            c = ((c & 1) ? (-306674912 ^ (c >>> 1)) : (c >>> 1));
            table[n] = c;
        }
        return typeof Int32Array !== 'undefined' ? new Int32Array(table) : table;
    }
    function crc32_buf(buf /*:ABuf*/, bufLen, seed /*:?CRC32Type*/) {
        if (!bufLen) {
            bufLen = buf.length;
        }
        if (bufLen > 10000)
            return crc32_buf_8(buf, bufLen, seed);
        var C = seed /*:: ? 0 : 0 */ ^ -1, L = bufLen - 3;
        for (var i = 0; i < L;) {
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
        }
        while (i < L + 3)
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
        return (C ^ (-1)) >>> 0;
    }
    Crc32.crc32_buf = crc32_buf;
    function crc32_buf_8(buf, bufLen, seed) {
        var C = seed ^ -1, L = bufLen - 7;
        for (var i = 0; i < L;) {
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
        }
        while (i < L + 7)
            C = (C >>> 8) ^ T[(C ^ buf[i++]) & 0xFF];
        return (C ^ (-1)) >>> 0;
    }
})(Crc32 = exports.Crc32 || (exports.Crc32 = {}));
var Protocol;
(function (Protocol) {
    /**
     * pomele client encode
     * id message id;
     * route message route
     * msg message body
     * socketio current support string
     */
    function strencode(str) {
        // encoding defaults to 'utf8'
        return buffer_1.Buffer.from(str);
    }
    Protocol.strencode = strencode;
    /**
     * client decode
     * msg String data
     * return Message Object
     */
    function strdecode(buffer) {
        // encoding defaults to 'utf8'
        return buffer.toString();
    }
    Protocol.strdecode = strdecode;
    function lzw_encode(s) {
        var dict = {};
        var data = (s + "").split("");
        var out = [];
        var currChar;
        var phrase = data[0];
        var code = 256;
        for (var i = 1; i < data.length; i++) {
            currChar = data[i];
            if (dict[phrase + currChar] != null) {
                phrase += currChar;
            }
            else {
                out.push(phrase.length > 1 ? dict[phrase] : phrase.charCodeAt(0));
                dict[phrase + currChar] = code;
                code++;
                phrase = currChar;
            }
        }
        out.push(phrase.length > 1 ? dict[phrase] : phrase.charCodeAt(0));
        for (var i = 0; i < out.length; i++) {
            out[i] = String.fromCharCode(out[i]);
        }
        return out.join("");
    }
    Protocol.lzw_encode = lzw_encode;
})(Protocol = exports.Protocol || (exports.Protocol = {}));
var Package;
(function (Package) {
    Package.TYPE_HANDSHAKE = 1;
    Package.TYPE_HANDSHAKE_ACK = 2;
    Package.TYPE_HEARTBEAT = 3;
    Package.TYPE_DATA = 4;
    Package.TYPE_KICK = 5;
    /**
     * Package protocol encode.
     *
     * Pinus package format:
     * +------+-------------+------------------+-------+------+
     * | type | body length |       body       | msgid | hash |
     * +------+-------------+------------------+-------+------+
     *
     * Head: 4bytes
     *   0: package type,
     *      1 - handshake,
     *      2 - handshake ack,
     *      3 - heartbeat,
     *      4 - data
     *      5 - kick
     *   1 - 3: big-endian body length
     *   Body: body length bytes
     *   1 - 4: msgId
     *   hash crc32
     * @param  {Number}    type   package type
     * @param  {Buffer} body   body content in bytes
     * @return {Buffer}        new byte array that contains encode result
     */
    function encode(type, body, nextSeqId) {
        var buffer;
        var length = body ? body.length : 0;
        var bufferLength = PKG_HEAD_BYTES + length;
        if (hasSeqIdEncode) {
            bufferLength += HASH_BYTES + END_MSGID_BYTES;
        }
        buffer = buffer_1.Buffer.alloc(bufferLength);
        var index = 0;
        buffer[index++] = type & 0xff;
        buffer[index++] = (length >> 16) & 0xff;
        buffer[index++] = (length >> 8) & 0xff;
        buffer[index++] = length & 0xff;
        if (body) {
            copyArray(buffer, index, body, 0, length);
        }
        if (hasSeqIdEncode) { //增加msgId
            var seqId = nextSeqId;
            index += length;
            buffer[index++] = (seqId >> 24) & 0xff;
            buffer[index++] = (seqId >> 16) & 0xff;
            buffer[index++] = (seqId >> 8) & 0xff;
            buffer[index++] = seqId & 0xff;
            var crc = Crc32.crc32_buf(buffer, index, SEED);
            buffer[index++] = (crc >> 24) & 0xff;
            buffer[index++] = (crc >> 16) & 0xff;
            buffer[index++] = (crc >> 8) & 0xff;
            buffer[index++] = crc & 0xff;
        }
        return buffer;
    }
    Package.encode = encode;
    function validate(type) {
        if ((Package.TYPE_HANDSHAKE === type ||
            Package.TYPE_HANDSHAKE_ACK === type ||
            Package.TYPE_HEARTBEAT === type ||
            Package.TYPE_DATA === type)) {
            return true;
        }
        return false;
    }
    Package.validate = validate;
    /**
     * Package protocol decode.
     * See encode for package format.
     *
     * @param  {Buffer} buffer byte array containing package content
     * @return {Object}           {type: package type, buffer: body byte array}
     */
    function decode(buffer, nextSeqId) {
        var offset = 0;
        var bytes = buffer_1.Buffer.from(buffer);
        var seqId = undefined;
        var length = 0;
        var checkOk = true;
        var crcHash = 0;
        var rs = [];
        var tempMsgStart = 0;
        while (offset < bytes.length) {
            tempMsgStart = offset;
            var type = bytes[offset++];
            if (!validate(type)) {
                return { 'type': type, checkOk: false };
            }
            // bufferPacket. bytes.subarray(offset, offset + 3)
            length = ((bytes[offset++]) << 16 | (bytes[offset++]) << 8 | bytes[offset++]) >>> 0;
            var bodyStart = offset;
            offset += length;
            if (hasSeqIdDecode) {
                // let bufferSeqId = bytes.subarray(offset, offset + 4)
                seqId = ((bytes[offset++]) << 24 | (bytes[offset++]) << 16 | (bytes[offset++]) << 8 | bytes[offset++]) >>> 0;
                if (nextSeqId !== undefined) {
                    if (seqId !== nextSeqId) {
                        return { 'type': type, checkOk: false };
                    }
                    var packetLen = 8 + length;
                    var buffPacket = new buffer_1.Buffer(packetLen);
                    copyArray(buffPacket, 0, bytes, tempMsgStart, packetLen);
                    var crcRet = Crc32.crc32_buf(buffPacket, offset, SEED);
                    crcHash = ((bytes[offset++]) << 24 | (bytes[offset++]) << 16 | (bytes[offset++]) << 8 | bytes[offset++]) >>> 0; //增加msgId
                    checkOk = (crcHash === crcRet);
                    if (!checkOk) {
                        return { 'type': type, checkOk: false };
                    }
                }
            }
            var body = length ? buffer_1.Buffer.alloc(length) : null;
            if (body) {
                copyArray(body, 0, bytes, bodyStart, length);
            }
            rs.push({ type: type, seqId: seqId, body: body, checkOk: checkOk }); //增加msgId
        }
        return rs.length === 1 ? rs[0] : rs;
    }
    Package.decode = decode;
})(Package = exports.Package || (exports.Package = {}));
var Message;
(function (Message) {
    Message.TYPE_REQUEST = 0;
    Message.TYPE_NOTIFY = 1;
    Message.TYPE_RESPONSE = 2;
    Message.TYPE_PUSH = 3;
    /**
     * Message protocol encode.
     *
     * @param  {Number} id            message id
     * @param  {Number} type          message type
     * @param  {Number} compressRoute whether compress route
     * @param  {Number|String} route  route code or route string
     * @param  {Buffer} msg           message body bytes
     * @return {Buffer}               encode result
     */
    function encode(id, type, compressRoute, route, msg, compressGzip) {
        // caculate message max length
        var idBytes = msgHasId(type) ? caculateMsgIdBytes(id) : 0;
        var msgLen = MSG_FLAG_BYTES + idBytes;
        if (msgHasRoute(type)) {
            if (compressRoute) {
                if (typeof route !== 'number') {
                    throw new Error('error flag for number route!');
                }
                msgLen += MSG_ROUTE_CODE_BYTES;
            }
            else {
                msgLen += MSG_ROUTE_LEN_BYTES;
                if (route) {
                    route = Protocol.strencode(route);
                    if (route.length > 255) {
                        throw new Error('route maxlength is overflow');
                    }
                    msgLen += route.length;
                }
            }
        }
        if (msg) {
            msgLen += msg.length;
        }
        var buffer = buffer_1.Buffer.alloc(msgLen);
        var offset = 0;
        // add flag
        offset = encodeMsgFlag(type, compressRoute, buffer, offset, compressGzip);
        // add message id
        if (msgHasId(type)) {
            offset = encodeMsgId(id, buffer, offset);
        }
        // add route
        if (msgHasRoute(type)) {
            offset = encodeMsgRoute(compressRoute, route, buffer, offset);
        }
        // add body
        if (msg) {
            offset = encodeMsgBody(msg, buffer, offset);
        }
        return buffer;
    }
    Message.encode = encode;
    /**
     * Message protocol decode.
     *
     * @param  {Buffer|Uint8Array} buffer message bytes
     * @return {Object}            message object
     */
    function decode(buffer) {
        var bytes = buffer_1.Buffer.from(buffer);
        var bytesLen = bytes.length || bytes.byteLength;
        var offset = 0;
        var id = 0;
        var route = null;
        // parse flag
        var flag = bytes[offset++];
        var compressRoute = flag & MSG_COMPRESS_ROUTE_MASK;
        var type = (flag >> 1) & MSG_TYPE_MASK;
        var compressGzip = (flag >> 4) & MSG_COMPRESS_GZIP_MASK;
        // parse id
        if (msgHasId(type)) {
            var m = 0;
            var i = 0;
            do {
                m = parseInt((bytes[offset] + ""));
                id += (m & 0x7f) << (7 * i);
                offset++;
                i++;
            } while (m >= 128);
        }
        // parse route
        if (msgHasRoute(type)) {
            if (compressRoute) {
                route = (bytes[offset++]) << 8 | bytes[offset++];
            }
            else {
                var routeLen = bytes[offset++];
                if (routeLen) {
                    route = buffer_1.Buffer.alloc(routeLen);
                    copyArray(route, 0, bytes, offset, routeLen);
                    route = Protocol.strdecode(route);
                }
                else {
                    route = '';
                }
                offset += routeLen;
            }
        }
        // parse body
        var bodyLen = bytesLen - offset;
        var body = buffer_1.Buffer.alloc(bodyLen);
        copyArray(body, 0, bytes, offset, bodyLen);
        return {
            'id': id, 'type': type, 'compressRoute': compressRoute,
            'route': route, 'body': body, 'compressGzip': compressGzip
        };
    }
    Message.decode = decode;
})(Message = exports.Message || (exports.Message = {}));
var copyArray = function (dest, doffset, src, soffset, length) {
    if ('function' === typeof src.copy) {
        // Buffer
        src.copy(dest, doffset, soffset, soffset + length);
    }
    else {
        // Uint8Array
        for (var index = 0; index < length; index++) {
            dest[doffset++] = src[soffset++];
        }
    }
};
var msgHasId = function (type) {
    return type === Message.TYPE_REQUEST || type === Message.TYPE_RESPONSE;
};
var msgHasRoute = function (type) {
    return type === Message.TYPE_REQUEST || type === Message.TYPE_NOTIFY ||
        type === Message.TYPE_PUSH;
};
var caculateMsgIdBytes = function (id) {
    var len = 0;
    do {
        len += 1;
        id >>= 7;
    } while (id > 0);
    return len;
};
var encodeMsgFlag = function (type, compressRoute, buffer, offset, compressGzip) {
    if (type !== Message.TYPE_REQUEST && type !== Message.TYPE_NOTIFY &&
        type !== Message.TYPE_RESPONSE && type !== Message.TYPE_PUSH) {
        throw new Error('unkonw message type: ' + type);
    }
    buffer[offset] = (type << 1) | (compressRoute ? 1 : 0);
    if (compressGzip) {
        buffer[offset] = buffer[offset] | MSG_COMPRESS_GZIP_ENCODE_MASK;
    }
    return offset + MSG_FLAG_BYTES;
};
var encodeMsgId = function (id, buffer, offset) {
    do {
        var tmp = id % 128;
        var next = Math.floor(id / 128);
        if (next !== 0) {
            tmp = tmp + 128;
        }
        buffer[offset++] = tmp;
        id = next;
    } while (id !== 0);
    return offset;
};
var encodeMsgRoute = function (compressRoute, _route, buffer, offset) {
    if (compressRoute) {
        var route = _route;
        if (route > MSG_ROUTE_CODE_MAX) {
            throw new Error('route number is overflow');
        }
        buffer[offset++] = (route >> 8) & 0xff;
        buffer[offset++] = route & 0xff;
    }
    else {
        var route = _route;
        if (route) {
            buffer[offset++] = route.length & 0xff;
            copyArray(buffer, offset, route, 0, route.length);
            offset += route.length;
        }
        else {
            buffer[offset++] = 0;
        }
    }
    return offset;
};
var encodeMsgBody = function (msg, buffer, offset) {
    copyArray(buffer, offset, msg, 0, msg.length);
    return offset + msg.length;
};

cc._RF.pop();