"use strict";
cc._RF.push(module, 'c1f3deLZSxEyaa4Xq9YbiRn', 'protobuf');
// Script/pomelo/protobuf/protobuf.ts

"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var buffer_1 = require("../buffer/buffer");
var decoder_1 = require("./decoder");
var encoder_1 = require("./encoder");
var parser = require("./parser");
var Protobuf = /** @class */ (function () {
    function Protobuf(opts) {
        // On the serverside, use serverProtos to encode messages send to client
        this.encoder = new encoder_1.Encoder(opts.encoderProtos);
        // On the serverside, user clientProtos to decode messages receive from clients
        this.decoder = new decoder_1.Decoder(opts.decoderProtos);
    }
    /**
     * [encode the given message, return a Buffer represent the message encoded by protobuf]
     * @param  {[type]} key The key to identify the message type.
     * @param  {[type]} msg The message body, a js object.
     * @return {[type]} The binary encode result in a Buffer.
     */
    Protobuf.prototype.encode = function (key, msg) {
        return this.encoder.encode(key, msg);
    };
    Protobuf.prototype.encode2Bytes = function (key, msg) {
        var buffer = this.encode(key, msg);
        if (!buffer || !buffer.length) {
            console.warn('encode msg failed! key : %j, msg : %j', key, msg);
            return null;
        }
        var bytes = new Uint8Array(buffer.length);
        for (var offset = 0; offset < buffer.length; offset++) {
            bytes[offset] = buffer.readUInt8(offset);
        }
        return bytes;
    };
    Protobuf.prototype.encodeStr = function (key, msg, code) {
        code = code || 'base64';
        var buffer = this.encode(key, msg);
        return !!buffer ? buffer.toString(code) : buffer;
    };
    Protobuf.prototype.decode = function (key, msg) {
        return this.decoder.decode(key, msg);
    };
    Protobuf.prototype.decodeStr = function (key, str, code) {
        code = code || 'base64';
        var buffer = buffer_1.Buffer.from(str, code);
        return !!buffer ? this.decode(key, buffer) : buffer;
    };
    Protobuf.parse = function (json) {
        return parser.parse(json);
    };
    Protobuf.prototype.setEncoderProtos = function (protos) {
        this.encoder.init(protos);
    };
    Protobuf.prototype.setDecoderProtos = function (protos) {
        this.decoder.init(protos);
    };
    return Protobuf;
}());
exports.Protobuf = Protobuf;

cc._RF.pop();