"use strict";
cc._RF.push(module, 'e7effOT8DhCsIRV7WAYHkmR', 'encoder');
// Script/pomelo/protobuf/encoder.ts

"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var buffer_1 = require("../buffer/buffer");
var codec = require("./codec");
var constant = require("./constant");
var util = require("./protobufutil");
var Encoder = /** @class */ (function () {
    function Encoder(protos) {
        this.init(protos);
    }
    Encoder.prototype.init = function (protos) {
        this.protos = protos || {};
    };
    Encoder.prototype.encode = function (route, msg) {
        if (!route || !msg) {
            console.warn('Route or msg can not be null! route : %j, msg %j', route, msg);
            return null;
        }
        // Get protos from protos map use the route as key
        var protos = this.protos[route];
        // Check msg
        if (!this.checkMsg(msg, protos)) {
            console.error('check msg failed! msg : %j, proto : %j', msg, protos);
            return null;
        }
        // Set the length of the buffer 2 times bigger to prevent overflow
        var length = buffer_1.Buffer.byteLength(JSON.stringify(msg)) * 2;
        // Init buffer and offset
        var buffer = buffer_1.Buffer.alloc(length);
        var offset = 0;
        if (!!protos) {
            offset = this.encodeMsg(buffer, offset, protos, msg);
            if (offset > 0) {
                return buffer.slice(0, offset);
            }
        }
        return null;
    };
    /**
     * Check if the msg follow the defination in the protos
     */
    Encoder.prototype.checkMsg = function (msg, protos) {
        if (!protos || !msg) {
            console.warn('no protos or msg exist! msg : %j, protos : %j', msg, protos);
            return false;
        }
        for (var name in protos) {
            var proto = protos[name];
            // All required element must exist
            switch (proto.option) {
                case 'required':
                    if (typeof (msg[name]) === 'undefined') {
                        console.warn('no property exist for required! name: %j, proto: %j, msg: %j', name, proto, msg);
                        return false;
                    }
                case 'optional':
                    if (typeof (msg[name]) !== 'undefined') {
                        var message_1 = protos.__messages[proto.type] || this.protos['message ' + proto.type];
                        if (!!message_1 && !this.checkMsg(msg[name], message_1)) {
                            console.warn('inner proto error! name: %j, proto: %j, msg: %j', name, proto, msg);
                            return false;
                        }
                    }
                    break;
                case 'repeated':
                    // Check nest message in repeated elements
                    var message = protos.__messages[proto.type] || this.protos['message ' + proto.type];
                    if (!!msg[name] && !!message) {
                        for (var i = 0; i < msg[name].length; i++) {
                            if (!this.checkMsg(msg[name][i], message)) {
                                return false;
                            }
                        }
                    }
                    break;
            }
        }
        return true;
    };
    Encoder.prototype.encodeMsg = function (buffer, offset, protos, msg) {
        for (var name in msg) {
            if (!!protos[name]) {
                var proto = protos[name];
                switch (proto.option) {
                    case 'required':
                    case 'optional':
                        offset = this.writeBytes(buffer, offset, this.encodeTag(proto.type, proto.tag));
                        offset = this.encodeProp(msg[name], proto.type, offset, buffer, protos);
                        break;
                    case 'repeated':
                        if (!!msg[name] && msg[name].length > 0) {
                            offset = this.encodeArray(msg[name], proto, offset, buffer, protos);
                        }
                        break;
                }
            }
        }
        return offset;
    };
    Encoder.prototype.encodeProp = function (value, type, offset, buffer, protos) {
        var length = 0;
        switch (type) {
            case 'uInt32':
                offset = this.writeBytes(buffer, offset, codec.encodeUInt32(value));
                break;
            case 'int32':
            case 'sInt32':
                offset = this.writeBytes(buffer, offset, codec.encodeSInt32(value));
                break;
            case 'float':
                buffer.writeFloatLE(value, offset);
                offset += 4;
                break;
            case 'double':
                buffer.writeDoubleLE(value, offset);
                offset += 8;
                break;
            case 'string':
                length = buffer_1.Buffer.byteLength(value);
                // Encode length
                offset = this.writeBytes(buffer, offset, codec.encodeUInt32(length));
                // write string
                buffer.write(value, offset, length);
                offset += length;
                break;
            default:
                var message = protos.__messages[type] || this.protos['message ' + type];
                if (!!message) {
                    // Use a tmp buffer to build an internal msg
                    var tmpBuffer = buffer_1.Buffer.alloc(buffer_1.Buffer.byteLength(JSON.stringify(value)) * 2);
                    length = 0;
                    length = this.encodeMsg(tmpBuffer, length, message, value);
                    // Encode length
                    offset = this.writeBytes(buffer, offset, codec.encodeUInt32(length));
                    // contact the object
                    tmpBuffer.copy(buffer, offset, 0, length);
                    offset += length;
                }
                break;
        }
        return offset;
    };
    /**
     * Encode reapeated properties, simple msg and object are decode differented
     */
    Encoder.prototype.encodeArray = function (array, proto, offset, buffer, protos) {
        var i = 0;
        if (util.isSimpleType(proto.type)) {
            offset = this.writeBytes(buffer, offset, this.encodeTag(proto.type, proto.tag));
            offset = this.writeBytes(buffer, offset, codec.encodeUInt32(array.length));
            for (i = 0; i < array.length; i++) {
                offset = this.encodeProp(array[i], proto.type, offset, buffer);
            }
        }
        else {
            for (i = 0; i < array.length; i++) {
                offset = this.writeBytes(buffer, offset, this.encodeTag(proto.type, proto.tag));
                offset = this.encodeProp(array[i], proto.type, offset, buffer, protos);
            }
        }
        return offset;
    };
    Encoder.prototype.writeBytes = function (buffer, offset, bytes) {
        for (var i = 0; i < bytes.length; i++) {
            buffer.writeUInt8(bytes[i], offset);
            offset++;
        }
        return offset;
    };
    Encoder.prototype.encodeTag = function (type, tag) {
        var value = constant.TYPES[type];
        if (value === undefined)
            value = 2;
        return codec.encodeUInt32((tag << 3) | value);
    };
    return Encoder;
}());
exports.Encoder = Encoder;

cc._RF.pop();